/*
//
// S-38.3610 Network Programming (Spring 2013)
// Assignment 2: HTTP Server
// Filip Nybck (79808K)
//
*/

//standard C headers
#include <errno.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

//unix headers
//#include <limits.h>
#include <signal.h>
#include <sys/stat.h>
#include <unistd.h>

//network headers
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

//local headers
#include "config.h"
#include "log.h"
#include "parse.h"
#include "string.h"
#include "transfer.h"
#include "threads.h"
#include "discovery.h"
#include "directory.h"

volatile bool run_server = true;

/**
 * Convert an IP address inside a sockaddr structure to a C string.
 *
 * @param sa The socket address containing the IP address to covert.
 * @param buffer The buffer to write the human readable address to.
 * @param buffer_length Length of the buffer in bytes.
 * @return Returns true on success, false on any failure.
 */
bool sockaddr_to_string(const struct sockaddr_storage *sa, char *buffer, 
	size_t buffer_length)
{
	const char *p;
 
 	if(sa == NULL)
 		return false;
 
  	if(sa->ss_family == AF_INET) {
 		p = inet_ntop(AF_INET, &((struct sockaddr_in *)sa)->sin_addr,
 			buffer, buffer_length);
	} else if(sa->ss_family == AF_INET6) {
		p = inet_ntop(AF_INET6, &((struct sockaddr_in6 *)sa)->sin6_addr,
 			buffer, buffer_length);
 	} else {
 		p = NULL;
 	}
 	
 	return p != NULL;
}

void print_usage()
{
	fprintf(stderr, "Use: %s [options ...]\n"
		"Options:\n"
		"--foreground\n"
		"\tThe server is not daemonized and writes log messages to stdout and "
			"stderr (in addition to the log file).\n"
		"--port port-number\n"
		"\tSets the port number the server is listening to. The same port "
			"number is used for both IPv4 and IPv6. The default port number is "
			"54321.\n"
		"-- root root-directory\n"
		"\tSets the root directory for the file storage. Files are got from "
			"and put to this directory. The default root directory is the "
			"files directory inside the current directory.\n"
		"--ipv4\n"
		"--ipv6\n"
		"\tListen only to connections over IPv4 or IPv6. If none or both of "
			"the --ipv4 and --ipv6 options are given, the server listens to "
			"connections over both protocols.\n"
		"--no-discovery\n"
		"\tDo not announce the server to the discovery service when starting "
			"and shutting down.\n"
		"--threads max-threads\n"
		"\tSets the maximum number of concurrent threads.\n",
		"--backlog backlog-length\n"
		"\tSets the length of the backlog queue of connections.\n"
		PROGRAM_NAME);
}

bool check_ul_option(int argi, int argc, char *argv[], const char *option_name, 
	unsigned long *value, unsigned long min, unsigned long max)
{
	unsigned long v;
	
	if(argv == NULL || option_name == NULL || value == NULL)
		return false;
	
	if(argi == argc) {
		log_printf(LL_ERROR, "Number missing after the %s option.", option_name);
		print_usage();
		return false;
	}
	
	errno = 0;
	v = strtoul(argv[argi], NULL, 10);
	if(errno != 0) {
		log_printf(LL_ERROR, "Number malformed after the %s option.", option_name);
		return false;
	}				
	
	if(v < min || v > max) {
		log_printf(LL_ERROR, "Number out of valid range [%lu,%lu] after the %s option.", 
			option_name, min, max);
		return false;
	}
	
	*value = v;
	
	return true;
}

/**
 * Check the arguments given when the program was started.
 *
 * @param argc Number of argument strings (from the main function).
 * @param argv Array of argument string pointers (from the main function).
 * @param config Configuration data to set.
 * @return Number of first non-option argument or -1 on failure.
 */
int check_options(int argc, char *argv[], struct CONFIG *config)
{
	int i;
	static char *default_root = DEFAULT_ROOT;
	unsigned long value;

	if(config == NULL)
		return -1;
	
	config->backlog_length = 10;
	config->port = 54321;
	config->root = default_root;
	config->ipv4_only = false;
	config->ipv6_only = false;
	config->daemonize = true;
	config->discovery = true;
	config->max_threads = 5;
	config->server_sock_fd = -1;

	for(i = 1; i < argc; i++) {
		if(strcmp(argv[i], "--port") == 0) {
			++i;
			if(!check_ul_option(i, argc, argv, "--port", &value, 1, 65535))
				return -1;
			config->port = value;			
		} else if(strcmp(argv[i], "--root") == 0) {
			i++;
			if(i == argc) {
				log_printf(LL_ERROR, "Content root directory missing after --root option.");
				print_usage();
				return -1;
			}
			config->root = argv[i];
		} else if(strcmp(argv[i], "--ipv4") == 0) {
			config->ipv4_only = true;
		} else if(strcmp(argv[i], "--ipv6") == 0) {
			config->ipv6_only = true;
		} else if(strcmp(argv[i], "--foreground") == 0) {
			config->daemonize = false;
		} else if(strcmp(argv[i], "--no-discovery") == 0) {
			config->discovery = false;
		} else if(strcmp(argv[i], "--threads") == 0) {
			i++;
			if(!check_ul_option(i, argc, argv, "--threads", &value, 1, 1000))
				return -1;
			config->max_threads = value;			
		} else if(strcmp(argv[i], "--backlog") == 0) {
			i++;
			if(!check_ul_option(i, argc, argv, "--backlog", &value, 1, 1000))
				return -1;
			config->backlog_length = value;			
		} else {
			log_printf(LL_ERROR, "Unknown command line option '%s'.", argv[i]);
			print_usage();
			return -1;
		}
	}
	
	if(config->ipv4_only && config->ipv6_only) {
		config->ipv4_only = false;
		config->ipv6_only = false;
	}
	
	return i;
}

/**
 * Create a local filename by concatenating the root directory with the
 * request path. The request path is relative to the root path.
 *
 * @param config The configuration structure including the root path, which
 *   must be a directory.
 * @param request_path The request path as given in the request.
 * @return Returns a concatenated string of the HTTP content root directory
 *   and the request path.
 */
char *get_local_filename(struct CONFIG *config, const char *request_path)
{
	size_t length;
	char *local_filename;

	if(config == NULL || request_path == NULL)
		return NULL;
	
	length = strlen(config->root) + strlen(request_path) + 1;
	local_filename = (char *)safer_malloc(length);
	if(local_filename == NULL)
		return NULL;
	
	strcpy(local_filename, config->root);
	strcat(local_filename, request_path);
	
	return local_filename;
}

/**
 * Map an internal status code (enum STATUS) to a HTTP status code, reason 
 * phrase and possibly a message body.
 *
 * @param status An internal status code.
 * @return Returns struct containing a HTTP status code, a reason phrase and 
 *   a message corresponding to the status code.
 */
struct HTTP_STATUS get_http_status(enum STATUS status)
{
	//see config.h for the macro containing the actual data
	static struct HTTP_STATUS http_status[] = HTTP_STATUS_DATA;
	struct HTTP_STATUS unknown_http_status = {000, "Unknown Error"};
	
	if(status < ARRAY_LENGTH(http_status))
		return http_status[status];
	
	unknown_http_status.status_code = status;
	return unknown_http_status;
}

/**
 * Send the reply header including the status line, the header lines and
 * the empty line at the end of the header.
 *
 * @param socket_stream The stream representing the HTTP connection to the 
 *   client.
 * @param reply_header The header structure containing the values to be used
 *   in the status line and the header lines.
 * @return Returns true on success, false on any failure. *
 */
bool send_header(FILE *socket_stream, struct HEADER *reply_header)
{
	bool ok = false;
	char buffer[BUFFERSIZE];
	struct tm time_struct;
	struct HTTP_STATUS http_status;

	if(socket_stream == NULL || reply_header == NULL)
		return false;
	
	//lookup HTTP status code, reason phrase and possibly body message
	http_status = get_http_status(reply_header->status);
	
	do {
		//log reply status line
		log_printf(LL_INFO, "HTTP/1.1 %03u %s", 
			http_status.status_code,
			http_status.reason_phrase);

		if(fprintf(socket_stream, 
			"HTTP/1.1 %03u %s\r\n"
			"Content-Type: %s\r\n"
			"Content-Length: %llu\r\n"
			"Iam: %s\r\n"
			"Connection: close\r\n",
			http_status.status_code,
			http_status.reason_phrase,
			reply_header->content_type,
			reply_header->content_length,
			IMPLEMENTOR_NAME) < 0) {
			break;
		}

		//optional header fields
		//currently Last-Modified
		if(reply_header->last_modified != 0) {
			gmtime_r(&reply_header->last_modified, &time_struct);
			strftime(buffer, sizeof(buffer), "%a, %d %b %Y %T GMT", &time_struct);
			if(fprintf(socket_stream, "Last-Modified: %s\r\n", buffer) < 0) {
				break;
			}
		}

		if(reply_header->status == STATUS_METHOD_NOT_ALLOWED) {
			if(fprintf(socket_stream, 
				"Allowed: GET\r\n") < 0) {
				break;
			}
		}

		//final empty line of the header
		if(fprintf(socket_stream, "\r\n") < 0) {
			break;
		}

		ok = true;
	} while(false);

	return ok;
}

/*
 * Send an error reply (header and possibly body) based on the status code 
 * provided.
 *
 * @param socket_stream The stream representing the client socket.
 * @param status_code Status code to report in the status line. This also 
 *   chooses the reason phrase in the status line and in the body.
 * @param method The requested method, currently either METHOD_GET or 
 *   METHOD_PUT. If the method is METHOD_GET, the error message is also sent 
 *   in the reply body so it is shown in web browsers.
 * @return Returns true on success, false on failure.
 */
bool send_error_reply(FILE *socket_stream, enum STATUS status, 
	enum METHOD method)
{
	struct HEADER *reply_header;
	bool ok = false;
	bool reply_body;
	char buffer[BUFFERSIZE];
	int length;
	struct HTTP_STATUS http_status;
	
	//If this is a reply to a GET request, send a reply body.
	reply_body = (method == METHOD_GET);

	reply_header = alloc_and_init_header(HEADER_TYPE_REPLY);
	if(reply_header == NULL) {
		log_printf(LL_ERROR, "reply_header == NULL in send_error_reply.");
		return false;
	}

	do {
		http_status = get_http_status(status);
	
		reply_header->status = status;
		reply_header->content_type = strdup("text/plain");
		reply_header->content_length = 0;

		//prepare the reply body
		if(reply_body) {
			//Create the body content. Could be something fancier, like HTML.
			//Currently status code and reason phrase.
			length = snprintf(buffer, sizeof(buffer), 
				"%03u %s\n",
				http_status.status_code, 
				http_status.body_message);
			if(length >= 0)
				reply_header->content_length = length;
		}

		if(!send_header(socket_stream, reply_header)) {
			log_printf(LL_ERRNO, "Error sending reply header.");
			break;
		}

		if(reply_body) {
			//send the body as binary
			if(fwrite(buffer, 1, reply_header->content_length, 
				socket_stream) < reply_header->content_length) {
				log_printf(LL_ERRNO, "Error sending reply body.");
				break;
			}
		}

		ok = true;
	} while(false);
	
	free_header(reply_header);
	
	return ok;
}

/**
 * Serve a GET request. In case something goes wrong and the connection to
 * the client is working, an error reply is sent.
 *
 * @param socket_stream The stream representing the HTTP connection to the 
 *   client.
 * @param local_filename The filename of the local file to send.
 * @param request_header The header struct from the parsed request.
 * @return Returns true on success, false on any failure.
 */
bool serve_get(FILE *socket_stream, const char *local_filename,
	struct HEADER *request_header)
{
	FILE *local_file = NULL;
	struct HEADER *reply_header = NULL;
	bool ok = false;
	struct stat local_file_stat;
	int local_file_fd;
	char *dir_list_buffer = NULL;

	do {
		reply_header = alloc_and_init_header(HEADER_TYPE_REPLY);
		if(reply_header == NULL) {
			log_printf(LL_ERROR, "Could not allocate and init reply header.");
			send_error_reply(socket_stream, STATUS_ERROR, METHOD_GET);
			return false;
		}
	
		//open local file
		local_file = fopen(local_filename, "rb");
		if(local_file == NULL) {
			log_printf(LL_ERRNO, "Error opening local file '%s' for reading.",
				local_filename);
			send_error_reply(socket_stream, STATUS_NOT_FOUND, METHOD_GET);
			break;
		}

		//fstat local file to get file size and modified time stamp
		local_file_fd = fileno(local_file);
		if(fstat(local_file_fd, &local_file_stat) != 0) {
			log_printf(LL_ERRNO, "Could not fstat local file '%s'.",
				local_filename);
			send_error_reply(socket_stream, STATUS_ERROR, METHOD_GET);
			break;
		}

		//prepare reply header
		reply_header->status = STATUS_OK;
		reply_header->content_type = strdup("text/plain");
		reply_header->content_length = local_file_stat.st_size;
		reply_header->last_modified = local_file_stat.st_mtime;
		
		//check if the file is a directory
		if(S_ISDIR(local_file_stat.st_mode)) {
//			log_printf(LL_ERROR, "Request to directory '%s'.", local_filename);
//			send_error_reply(socket_stream, STATUS_DIRECTORY, METHOD_GET);
			dir_list_buffer = (char *)safer_malloc(BUFFERSIZE);
			if(dir_list_buffer == NULL) {
				send_error_reply(socket_stream, STATUS_ERROR, METHOD_GET);
				break;
			}

			if(!list_directory(local_filename, reply_header, 
				dir_list_buffer, BUFFERSIZE))
			{
				send_error_reply(socket_stream, STATUS_ERROR, METHOD_GET);
				break;
			}
		}

		//send the header
		if(!send_header(socket_stream, reply_header)) {
			log_printf(LL_ERRNO, "Error sending reply header.");
			break;
		}
		
		//send the body
		//the body is file contents
		if(!S_ISDIR(local_file_stat.st_mode)) {
			if(!send_binary_data(reply_header->content_length, 
				socket_stream, local_file))
			{
				log_printf(LL_ERRNO, "Error sending reply body.");
				break;
			}
		}
		//the body is a directory listing
		else {
			if(!send_binary_mem_data(reply_header->content_length,
				socket_stream, dir_list_buffer))
			{
				log_printf(LL_ERRNO, "Error sending reply body.");
				break;
			}
		}
		
		ok = true;
	} while(false);

	//clean up	
	if(reply_header != NULL)
		free_header(reply_header);
	if(local_file != NULL)
		fclose(local_file);
	if(dir_list_buffer != NULL)
		safer_free(dir_list_buffer);
		
	return ok;
}

/**
 * Serve a PUT request. In case something goes wrong and the connection to
 * the client is working, an error reply is sent.
 *
 * @param socket_stream The stream representing the HTTP connection to the 
 *   client.
 * @param local_filename The filename of the local file to store to.
 * @param request_header The header struct from the parsed request.
 * @return Returns true on success, false on any failure.
 */
bool serve_put(FILE *socket_stream, const char *local_filename,
	struct HEADER *request_header)
{
	FILE *local_file = NULL;
	struct HEADER *reply_header = NULL;
	bool ok = false;
	struct stat local_file_stat;

	if(socket_stream == NULL || local_filename == NULL || request_header == NULL)
		return false;

	do {
		//check that a valid Content-Length is provided in the request header
		if(!request_header->content_length_valid) {
			log_printf(LL_ERROR, "Content-Length missing from the request header.");
			send_error_reply(socket_stream, STATUS_LENGTH_REQUIRED, METHOD_PUT);
			break;
		}
	
		//stat local file to check if this is a directory
		if(stat(local_filename, &local_file_stat) == 0) {
			if(S_ISDIR(local_file_stat.st_mode)) {
				log_printf(LL_ERRNO, "Cannot PUT to a directory '%s'.",
					local_filename);
				send_error_reply(socket_stream, STATUS_METHOD_NOT_ALLOWED, 
					METHOD_PUT);
				break;
			}
		}
		
		//open local file
		local_file = fopen(local_filename, "wb");
		if(local_file == NULL) {
			log_printf(LL_ERRNO, "Error opening local file '%s' for writing.",
				local_filename);
			send_error_reply(socket_stream, STATUS_NOT_FOUND, METHOD_PUT);
			break;
		}

		//receive the body
		if(!receive_binary_data(request_header->content_length, 
			socket_stream, local_file)) {
			log_printf(LL_ERRNO, "Error receiving request body.");
			break;
		}
		
		reply_header = alloc_and_init_header(HEADER_TYPE_REPLY);
		if(reply_header == NULL) {
			log_printf(LL_ERROR, "Could not allocate and init reply header.");
			send_error_reply(socket_stream, STATUS_ERROR, METHOD_PUT);
			break;
		}

		//prepare reply header
		reply_header->status = STATUS_OK;
		reply_header->content_type = strdup("text/plain");
		reply_header->content_length = 0;

		//send the reply header
		if(!send_header(socket_stream, reply_header)) {
			log_printf(LL_ERRNO, "Error sending reply header.");
			break;
		}
		
		ok = true;
	} while(false);

	//clean up	
	if(reply_header != NULL)
		free_header(reply_header);
	if(local_file != NULL)
		fclose(local_file);
		
	return ok;
}

/**
 * Serve a HTTP request from a client.
 *
 * @param config Configuration data to use.
 * @param client_sock_fd The file descriptor representing the HTTP connection 
 *   to the client.
 * @param client_sock_addr Socket address of the client for printing and 
 *   logging.
 * @return Returns true on success, false on any failure.
 */
bool serve_request(struct CONFIG *config, int client_sock_fd, 
	struct sockaddr_storage *client_sock_addr)
{
	FILE *client_sock;
	char addr_buffer[100];
	struct HEADER *request_header = NULL;
	enum STATUS status;
	bool ok = false;
	char *local_filename = NULL;
	
	//First log the client IP address and port.
	sockaddr_to_string(client_sock_addr, addr_buffer, sizeof(addr_buffer));
	log_printf(LL_INFO, "Client connected from %s at port %u.", addr_buffer, 
		ntohs(((struct sockaddr_in *)client_sock_addr)->sin_port));

	client_sock = fdopen(client_sock_fd, "r+b");
	if(client_sock == NULL) {
		log_printf(LL_ERRNO, "Could not open stream from socket descriptor in serve_request.");
		return false;
	}

	do {
		ok = false;
	
		//parse the header
		request_header = parse_header(client_sock, HEADER_TYPE_REQUEST, 
			&status);
		if(request_header == NULL || status != STATUS_OK) {
			//log_printf(LL_ERROR, "Error parsing request header (status: %u).",
			//	status);
			send_error_reply(client_sock, status, METHOD_GET);
			break;
		}

		//log request line
		log_printf(LL_INFO, "%s", request_header->request_line);

		//decode request path
		if(!decode_request_path(request_header->request_path)) {
			log_printf(LL_ERROR, "Request path not valid (decoding error): '%s'", 
				request_header->request_path);
			send_error_reply(client_sock, STATUS_SYNTAX_ERROR_IN_REQUEST_PATH, 
				METHOD_GET);
			break;
		}

		//validate request path
		if(!validate_request_path(request_header->request_path)) {
			log_printf(LL_ERROR, "Request path not valid: '%s'", 
				request_header->request_path);
			send_error_reply(client_sock, STATUS_SYNTAX_ERROR_IN_REQUEST_PATH, 
				METHOD_GET);
			break;
		}
		
		local_filename = get_local_filename(config, request_header->request_path);
		if(local_filename == NULL) {
			log_printf(LL_ERROR, "Could not create a local filename string "
				"from the root and request path.");
			send_error_reply(client_sock, STATUS_ERROR, METHOD_GET);
			break;
		}
		//log_printf(LL_DEBUG, "local filename: '%s'", local_filename);
		
		//dispatch to the method requested
		if(request_header->method == METHOD_GET) {
			if(!serve_get(client_sock, local_filename, request_header))
				break;
		} else if(request_header->method == METHOD_PUT) {
			if(!serve_put(client_sock, local_filename, request_header))
				break;
		}

		ok = true;
	} while(false);

	free_header(request_header);
	if(local_filename != NULL) {
		safer_free(local_filename);
		local_filename = NULL;
	}
	
	//Close the connection. We don't need any half duplex communication
	//that can be achieved with shutdown, but we want to close the connection
	//in both directions. The close call (indirectly called by fclose) takes
	//care of that. fclose also flushes the buffers before closing.
	fclose(client_sock);
	
	log_printf(LL_INFO, "Connection closed (%s at port %u).", addr_buffer, 
		ntohs(((struct sockaddr_in *)client_sock_addr)->sin_port));
	
	return ok;
}

/**
 * A wrapper function that is started by pthread_create and can only take one
 * (void *) parameter. This function calls the serve_request function with
 * arguments stored in the thread struct. Finally releases the thread struct.
 *
 * @param arg Pointer to a thread struct (allocated by acquire_thread_struct)
 *   cast as void * as required by pthread_create.
 * @return Always returns NULL, threads are not joined when finished.
 */
void *serve_request_wrapper(void *arg)
{
	struct THREAD *thread = (struct THREAD *)arg;

	serve_request(thread->args.config, 
		thread->args.client_sock_fd, 
		&thread->args.client_sock_addr);
	
	//Release the thread struct.
	release_thread_struct(thread);
	
	return NULL;
}

/**
 * Spawn a new thread in order to serve the HTTP request from a client.
 *
 * @param config Configuration data to use.
 * @param client_sock_fd The file descriptor representing the HTTP connection 
 *   to the client.
 * @param client_sock_addr Socket address of the client for printing and 
 *   logging.
 * @return Returns true on success, false on any failure.
 */
bool spawn_request_thread(struct CONFIG *config, int client_sock_fd, 
	struct sockaddr_storage *client_sock_addr)
{
	bool ok = false;
	struct THREAD *thread = NULL;
	
	do {
		//allocate a thread struct from the thread table
		thread = alloc_thread_struct(config);
		if(thread == NULL) {
			log_printf(LL_ERROR, "Cannot create more threads.");
			break;
		}

		//fill in the arguments to serve_request
		thread->args.config = config;
		thread->args.client_sock_fd = client_sock_fd;
		thread->args.client_sock_addr = *client_sock_addr;
	
		//spawn a new thread
		if(pthread_create(&thread->id, NULL, &serve_request_wrapper, 
			thread) != 0) {
			log_printf(LL_ERROR, "pthread_create failed.");
			break;
		}
	
		ok = true;
	} while(false);

	//cannot serve request, we have to reject it
	if(!ok) {
		release_thread_struct(thread);
		close(client_sock_fd);
	}

	return ok;	
}

/**
 * Set up the server (=listening) socket.
 *
 * @param config Configuration data to use.
 * @return Returns the server (=listening) socket on success, -1 on failure.
 */
int server_create_socket(struct CONFIG *config)
{
	int server_sock_fd;
	struct sockaddr_storage sock_addr;
	struct sockaddr_in *sock_addr_4 = (struct sockaddr_in *)&sock_addr;
	struct sockaddr_in6 *sock_addr_6 = (struct sockaddr_in6 *)&sock_addr;
	int value;
	bool ok = false;
	socklen_t sock_len;

	do {
		//clear the sockaddr struct
		memset(&sock_addr, 0, sizeof(sock_addr));
	
		//fill the socket address
		//no lookup needed, listen to any local IP
		if(config->ipv4_only) {
			sock_addr_4->sin_family = AF_INET;
			sock_addr_4->sin_port = htons(config->port);
			sock_addr_4->sin_addr.s_addr = INADDR_ANY;
			sock_len = sizeof(struct sockaddr_in);
		} else {
			sock_addr_6->sin6_family = AF_INET6;
			sock_addr_6->sin6_port = htons(config->port);
			sock_addr_6->sin6_addr = in6addr_any;
			sock_len = sizeof(struct sockaddr_in6);
		}
	
		server_sock_fd = socket(sock_addr.ss_family, SOCK_STREAM, 0);
		if(server_sock_fd == -1) {
			log_printf(LL_ERRNO, "Could not create listening socket.");
			break;
		}
		
		value = 1;
		if(setsockopt(server_sock_fd, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(value)) != 0) {
			log_printf(LL_ERRNO, "Could not set socket option REUSEADDR.");
			break;
		}
		
		if(config->ipv6_only) {
			value = 1;
			if(setsockopt(server_sock_fd, IPPROTO_IPV6, IPV6_V6ONLY, &value, sizeof(value)) != 0) {
				log_printf(LL_ERRNO, "Could not set socket option IPV6_V6ONLY.");
				break;
			}
		}
		
		if(bind(server_sock_fd, (struct sockaddr *)&sock_addr, sock_len) != 0) {
			log_printf(LL_ERRNO, "Could not bind socket address.");
			break;
		}	
		
		if(listen(server_sock_fd, config->backlog_length) != 0) {
			log_printf(LL_ERRNO, "Could not listen to socket address.");
			break;
		}

		config->server_sock_fd = server_sock_fd;

		ok = true;
	} while(false);

	if(!ok) {
		if(server_sock_fd != -1)
			close(server_sock_fd);
		server_sock_fd = -1;
	}

	return server_sock_fd;
}

/**
 * The main server loop. Wait for connections using accept.
 * When a connection has been made from a client, serve_request is called.
 * If and when the loop is exited (SIGINT is issued), the listening socket is 
 * closed.
 *
 * @param config Configuration data to use.
 * @return Returns false on any error, true otherwise.
 */
bool server_loop(struct CONFIG *config)
{
	int client_sock_fd;
	struct sockaddr_storage client_sock_addr;
	socklen_t client_sock_len;
	bool ok = true;
	
	if(config == NULL || config->server_sock_fd == -1)
		return false;
	
	while(run_server) {
		client_sock_len = sizeof(client_sock_addr);
	
		client_sock_fd = accept(config->server_sock_fd, 
			(struct sockaddr *)&client_sock_addr, 
			&client_sock_len);
		if(client_sock_fd == -1) {
			if(errno == EINTR) {
				break;
			}
		
			log_printf(LL_ERRNO, "Could not accept incoming connection.");
			ok = false;
			break;
		}

		spawn_request_thread(config, client_sock_fd, &client_sock_addr);
	}
	
	return ok;
}

/**
 * Daemonize the server process. Detach the terminal by forking and creating a 
 * new session. Redirect stdin, stdout and stderr to /dev/null and close any
 * other files.
 *
 * @return Returns true on success, false on failure.
 **/
bool daemonize()
{
	pid_t process_id, session_id;
	int fd, max_fd, log_fd;
	
	//fork
	process_id = fork();
	if(process_id == -1) {
		log_printf(LL_ERRNO, "fork failed.");
		return false;	
	} else if(process_id > 0) {
		//the parent process exits
		exit(EXIT_SUCCESS);
	}
	
	//the child process continues
	//create a new session
	session_id = setsid();
	if(session_id == -1) {
		log_printf(LL_ERRNO, "setsid failed.");
		return false;	
	}

	//this process is now the session leader
	//fork again, so that the new process is not a session leader
	process_id = fork();
	if(process_id == -1) {
		log_printf(LL_ERRNO, "fork failed.");
		return false;	
	} else if(process_id > 0) {
		//the parent process exits
		exit(EXIT_SUCCESS);
	}
	
	
	//redirect stdin, stdout, stderr (fds 0, 1 and 2) to /dev/null
	freopen("/dev/null", "r", stdin);
    freopen("/dev/null", "w", stdout);
    freopen("/dev/null", "w", stderr);

	//get the maximum number of file descriptors
	max_fd = sysconf(_SC_OPEN_MAX);
	log_printf(LL_DEBUG, "Max number of open files: %d.", max_fd);
	//get the file descriptor of the log file
	log_fd = log_get_file_descriptor();
	//Close every other file descriptor except for the log file.
	//If sysconf did not succeed and returned -1, no iterations will be
	//made in the for loop.
	for(fd = 3; fd < max_fd; fd++) {
		if(fd != log_fd)
			close(fd);
	}
	
	//the child process continues
	log_printf(LL_INFO, "The server has process id %d after daemonization.", getpid());

	return true;
}

/**
 * Signal handler function. Currently only the SIGINT signal is reacted upon
 * (e.g. when Ctrl-C is pressed).
 *
 * @param sig The signal that is issued.
 */
void signal_handler(int sig)
{
	if(sig == SIGINT || sig == SIGTERM) {
		run_server = false;
	}
}

/**
 * Set up signal handlers for the SIGINT (triggered by Ctrl-C from a terminal
 * if the server does not run as a daemon) and SIGTERM (the default signal
 * sent by the kill command. 
 *
 * @return Returns true on success, false on failure.
 */
bool init_signals()
{
	struct sigaction sa, sigpipe_sa;
	
	//for SIGINT and SIGTERM
	sa.sa_handler = &signal_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	//for SIGPIPE
	sigpipe_sa.sa_handler = SIG_IGN;
	sigemptyset(&sigpipe_sa.sa_mask);
	sigpipe_sa.sa_flags = 0;
	
	if(sigaction(SIGINT, &sa, NULL) != 0 
		|| sigaction(SIGTERM, &sa, NULL) != 0
		|| sigaction(SIGPIPE, &sigpipe_sa, NULL) != 0)
	{
		log_printf(LL_ERRNO, "Could not set signal handlers.");
		return false;
	}
	
	
	
	return true;
}

/**
 * The main function.
 *
 * @param argc Number of argument strings.
 * @param argv Array of argument string pointers.
 * @return Returns EXIT_SUCCESS on success or EXIT_FAILURE on failure.
 */
int main(int argc, char *argv[])
{
	struct CONFIG config;
	bool ok = false;
	
	do {
		if(check_options(argc, argv, &config) == -1) {
			return EXIT_FAILURE;
		}
	
	 	log_open(SERVER_LOG_FILENAME);
		//log_printf(LL_INFO, "Starting server.");

	 	if(config.daemonize) {
			if(!daemonize())
				break;
		}

		if(!init_signals())
			break;
		
		if(!init_threads(&config))
			break;

		//Create the server (=listening) socket.
		config.server_sock_fd = server_create_socket(&config);
		if(config.server_sock_fd == -1)
			break;

		if(config.discovery)
			discovery_service_join(&config);
		
		log_printf(LL_INFO, "Server started.");
		
		if(!server_loop(&config))
			break;

		ok = true;
	} while(false);
	
	log_printf(LL_INFO, "Shutting down server.");
	
	if(config.discovery)
		discovery_service_leave(&config);
	
	//close the server socket
	close(config.server_sock_fd);
	
	shutdown_threads();
	
	log_printf(LL_INFO, "Server shut down.");
	log_close();
	
	return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}

