/*
//
// S-38.3610 Network Programming (Spring 2013)
// Assignment 2: HTTP Server
// Filip Nybck (79808K)
//
*/

//standard C headers
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

//unix headers
#include <pthread.h>

//local headers
#include "config.h"
#include "log.h"

FILE *log_file = NULL;
//lock for the log to prevent concurrent threads to mix up the log
pthread_mutex_t log_mutex;


/**
 * Get current time as a string for logging.
 * Changes: localtime changed to localtime_r for thread safety.
 *
 * @param buffer The buffer to place the time string in.
 * @param buffer_size Size of the buffer.
 * @return Returns true on success, false on failure.
 */
bool log_time_string(char *buffer, size_t buffer_size)
{
	time_t timestamp;
	struct tm time_struct;

	timestamp = time(NULL);
	localtime_r(&timestamp, &time_struct);

	return strftime(buffer, buffer_size, "%Y-%m-%d %T", &time_struct) != 0;
}

/**
 * Open the log file for logging and initialize the log lock.
 *
 * @return Returns true on success, false on failure.
 */
bool log_open(const char *log_filename)
{
	//file mode is append mode
	log_file = fopen(log_filename, "a");
	
	if(log_file == NULL) {
		log_printf(LL_ERROR, "Could not open log file.");
		return false;
	}
	
	//create the log lock
	pthread_mutex_init(&log_mutex, NULL);

	return true;
}

/**
 * Close the log file.
 *
 * @return Returns true on success, false on failure.
 */
bool log_close()
{
	bool ok;

	ok = fclose(log_file) == 0;
	log_file = NULL;

	//destroy the log lock
	pthread_mutex_destroy(&log_mutex);

	return ok;
}

/**
 * Print a log record. The log message is printed to stdout or stderror,
 * depending o the log level. In case the server is daemonized, stdout and 
 * stderr are redirected to /dev/null and the output is lost. If a log file has 
 * been successfully opened, the log message is prepended with date and time
 * and written to the log file.
 *
 * @param level The log level of the message to be written. See log.h for log
 *   levels.
 * @param format_string Format string of the log message.
 * @return Returns true on success, false on failure.
 */
bool log_printf(enum LOG_LEVEL level, const char *format_string, ...)
{
	va_list arg_list;
	char time_string[100];
	FILE *output_stream;
	int ok = 1;

	if(level == LL_ERROR || level == LL_ERRNO)
		output_stream = stderr;
	else
		output_stream = stdout;

	if(!log_time_string(time_string, sizeof(time_string)))
		return false;

	//acquire the log lock
	pthread_mutex_lock(&log_mutex);

	if(log_file != NULL) {
		//print time stamp only to log file
		ok &= (fprintf(log_file, "%s ", time_string) >= 0);

		va_start(arg_list, format_string);
		ok &= (vfprintf(log_file, format_string, arg_list) >= 0);
		va_end(arg_list);
	}
	
	va_start(arg_list, format_string);
	ok &= (vfprintf(output_stream, format_string, arg_list) >= 0);
	va_end(arg_list);
		
	if(level == LL_ERRNO) {
		if(log_file != NULL)
			ok &= (fprintf(log_file, " Additional error info: %s ", 
				strerror(errno)) >= 0);
		ok &= (fprintf(output_stream, " Additional error info: %s ", 
				strerror(errno)) >= 0);
	}
	
	if(log_file != NULL) {
		ok &= (fprintf(log_file, "\n") >= 0);
		fflush(log_file);
	}
	ok &= (fprintf(output_stream, "\n") >= 0);
	fflush(output_stream);

	//release the log lock
	pthread_mutex_unlock(&log_mutex);

	return ok != 0;
}

/**
 * Get the file descriptor of the log file.
 *
 * @return Returns the file descriptor of the log file if the log file has been
 *   opened.
 */
int log_get_file_descriptor()
{
	return fileno(log_file);
}


