/*
//
// S-38.3610 Network Programming (Spring 2013)
// Assignment 2: HTTP Server
// Filip Nybck (79808K)
//
*/

#include <ctype.h>
#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "log.h"
#include "parse.h"
#include "pstring.h"
#include "string.h"
#include "threads.h"

/**
 * Read one line from the socket_stream into the buffer with the specified 
 * size. Remove the \r\n or \n newline sequence at the end of the line.
 *
 * @param buffer The buffer to read the line into.
 * @param buffer_size Size of the buffer.
 * @param socket_stream The socket stream to read from.
 * @return Returns the length of the string (not counting the line terminator)
 *   or < 0 in case of failure.
 */
size_t get_header_line(char *buffer, size_t buffer_size, FILE *socket_stream)
{
	char *s;
	size_t length;

	//According to RFC2616 (HTTP/1.1), a receiver should also accept a single 
	//\n as a line terminator. In other words, ignore the \r in the \r\n
	//sequence when receiving.
	
	s = fgets(buffer, buffer_size, socket_stream);
	if(s == NULL)
		return -1;
	
	//remove the \r\n or \n at the end of the line
	length = strlen(s);
	if(length >= 1 && s[length - 1] == '\n') {
		s[--length] = '\0';
		
		if(length >= 1 && s[length - 1] == '\r') {
			s[--length] = '\0';
		}
	}

	return length;
}

/**
 * Parse the status line. Get the HTTP version and status code.
 * 
 * @param s The status line to parse.
 * @param header The header structure to store the HTTP version and status code 
 *    in.
 * @return Returns true on success, false on failure.    
 */
bool parse_status_line(const char *s, struct HEADER *header)
{
	size_t len;

	//a status line has the form:
	//HTTP/x.y<ws>zzz<ws><reason phrase>
	//where x.y is the HTTP version
	//<ws> is at least one whitespace character
	//zzz is the three-digit status code
	//<reason phrase> is a textual explanation of the status code
	
	if(s == NULL || header == NULL)
		return false;
	
	if(strncmp(s, "HTTP/", 5) != 0)
		return false;
	s+= 5;

	//the HTTP version x.y
	//x and y can each be more than a single digit
	//see section 3.1 of the HTTP/1.1 RFC 2616
	len = strspn(s, DIGITS);
	if(len < 1)
		return false;
	header->http_version_major = strtoul(s, NULL, 10);
	s += len;
	
	if(*s++ != '.')
		return false;

	len = strspn(s, DIGITS);
	if(len < 1)
		return false;
	header->http_version_minor = strtoul(s, NULL, 10);
	s += len;

	//skip whitespace characters
	//at least one must be present
	//see section 19.3 of the HTTP/1.1 RFC 2616
	len = strspn(s, WHITESPACE);
	if(len < 1)
		return false;
	s += len;
	
	//the three-digit status code
	len = strspn(s, DIGITS);
	if(len != 3)
		return false;
	
	header->status_code = strndup(s, 3);
	s += 3;

	//ignore the reason phrase including the whitespace before it
/*	
	//skip whitespace characters
	//at least one must be present
	len = strspn(s, WHITESPACE);
	if(len < 1)
		return false;
	s += len;
*/	
	
	return true;	
}

/**
 * Validate and parse the request line and store the request method, request URL
 * and HTTP version in the provided header structure.
 *
 * @param s The C string representing the request line.
 * @param header The header structure to store the request method, request URL 
 *     and HTTP version in.
 * @return Returns 0 on success or a HTTP status code on failure.
 */
enum STATUS parse_request_line(const char *s, struct HEADER *header)
{
	size_t len;
	struct PSTRING method;

	//A request line has the form:
	//<method><ws><request-URL><ws>HTTP/x.y
	//<method> is the request method, either GET or PUT
	//<ws> is at least one whitespace character
	//<request-url> is the request URL of the requested resource
	//x.y is the HTTP version
	
	if(s == NULL || header == NULL)
		return STATUS_ERROR;
	
	//The method. At least one non-whitespace character must be present.
	//Check the method string later in this function.
	len = strcspn(s, WHITESPACE);
	if(len < 1)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	pstring_init(&method, s, len);
	s += len;
	
	//Skip whitespace characters. At least one must be present
	//(see section 19.3 of the HTTP/1.1 RFC 2616).
	len = strspn(s, WHITESPACE);
	if(len < 1)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	s += len;

	//The absolute path. At least one non-whitespace character must be present.
	//The absolute path is not further examined in this function.
	len = strcspn(s, WHITESPACE);
	if(len < 1)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	header->request_path = strndup(s, len);
	s += len;
	
	//Skip whitespace characters.
	len = strspn(s, WHITESPACE);
	if(len < 1)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	s += len;

	if(strncmp(s, "HTTP/", 5) != 0)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	s+= 5;

	//The HTTP version x.y,
	//where x and y can each be more than a single digit.
	//See section 3.1 of the HTTP/1.1 RFC 2616.
	len = strspn(s, DIGITS);
	if(len < 1)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	header->http_version_major = strtoul(s, NULL, 10);
	s += len;
	
	if(*s++ != '.')
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;

	len = strspn(s, DIGITS);
	if(len < 1)
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;
	header->http_version_minor = strtoul(s, NULL, 10);
	s += len;

	//Skip whitespace characters.
	len = strspn(s, WHITESPACE);
	s += len;

	//still something else after the HTTP version?
	if(*s != '\0')
		return STATUS_SYNTAX_ERROR_IN_REQUEST_LINE;

	//Check the method string. The method names are case sensitive.
	//See section 5.1.1 of the HTTP/1.1 RFC 2616.
	if(pstring_equals(&method, "GET"))
		header->method = METHOD_GET;
	else if(pstring_equals(&method, "PUT"))
		header->method = METHOD_PUT;
	else
		return STATUS_NOT_IMPLEMENTED;

	//Parsing the request line succeeded.
	return STATUS_OK;
}

/**
 * Parse one header line by isolating trimmed name and value substrings.
 *
 * @param s The header line to parse.
 * @param name The trimmed name substring.
 * @param value The trimmed value substring.
 * @return Returns true on success, false on failure.
 */
bool parse_header_line(const char *s, struct PSTRING *name, struct PSTRING *value)
{
	size_t len;
	
	//a header line has the form:
	//Name<ws>:<ws>Value<ws>
	//This imlementation does not support header lines spanning multiple 
	//lines as defined in section 4.2 of the HTTP/1.1 RFC 2616.
	
	if(s == NULL || name == NULL || value == NULL)
		return false;
	
	//get the name
	//the name must be at least one character
	len = strcspn(s, COLON_WHITESPACE);
	if(len < 1)
		return false;
	pstring_init(name, s, len);
	s += len;
	
	//skip whitespace, optional
	len = strspn(s, WHITESPACE);
	s += len;
	
	if(*s++ != ':')
		return false;

	//skip whitespace, optional
	len = strspn(s, WHITESPACE);
	s += len;
	
	//get the value
	//The rest of the line is the value. Trim whitespace from the end.
	len = strlen(s);
	while(len > 0 && (s[len - 1] == ' ' || s[len - 1] == '\t'))
		len--;
	
	pstring_init(value, s, len);
	
	return true;	
}

/**
 * Allocate a header structure and initialize its pointers to NULL.
 *
 * @param type The type of header to be created: either HEADER_TYPE_REQUEST
 *   or HEADER_TYPE_REPLY.
 * @return Returns the header structure on succes or NULL on failure.
 */
struct HEADER *alloc_and_init_header(enum HEADER_TYPE type)
{
	struct HEADER *header;

	//check parameters
	if(type != HEADER_TYPE_REQUEST && type != HEADER_TYPE_REPLY)
		return NULL;

	header = (struct HEADER *)safer_malloc(sizeof(struct HEADER));
	if(header == NULL)
		return NULL;
	
	header->type = type;
	header->request_line = NULL;
	header->request_path = NULL;
	header->status_code = NULL;
	header->status_line = NULL;
	header->content_length_valid = false;
	//header->content_length = 0;
	header->content_type = NULL;
	header->iam = NULL;
	header->last_modified = 0;
	
	return header;
}

/**
 * Receive and parse the status line or request line, depending on the header 
 * type. Receive header lines, parse them and fill the header structure with
 * relevant information.
 *
 * @param socket_stream The stream representing the TCP socket to read from.
 * @param type Header type: either HEADER_TYPE_REQUEST or HEADER_TYPE_REPLY.
 * @param status_code Optional output parameter to return the HTTP status
 *     code so far. Can be NULL.
 * @return Returns a parsed header on success, NULL on failure.
 */
struct HEADER *parse_header(FILE *socket_stream, enum HEADER_TYPE type,
	enum STATUS *status)
{
	int i;
	size_t length;
	struct HEADER *header;
	char buffer[BUFFERSIZE];
	bool ok = true;
	struct PSTRING name, value;
	enum STATUS local_status, t;
	
	//If status == NULL, we set the status pointer to the address of
	//a local dummy variable (local_status) that is never read. This way 
	//we always point to some valid address and don't have to check if 
	//status is NULL all the time we write to it.
	if(status == NULL)
		status = &local_status;
	*status = STATUS_OK;
	
	//check parameters
	if(socket_stream == NULL 
		|| (type != HEADER_TYPE_REQUEST && type != HEADER_TYPE_REPLY)) {
		*status = STATUS_ERROR;
		return NULL;
	}
	
	header = alloc_and_init_header(type);
	if(header == NULL) {
		*status = STATUS_ERROR;
		return NULL;
	}

	i = 0;
	do {
		
		errno = 0;
		length = get_header_line(buffer, sizeof(buffer), socket_stream);
		if(length < 0) {
			log_printf(LL_ERRNO, "Could not read header line.");
			ok = false;
			*status = STATUS_ERROR;
			break;
		}
		
		//request line
		if(i == 0 && type == HEADER_TYPE_REQUEST) {
			t = parse_request_line(buffer, header);
			if(t != STATUS_OK) {
				log_printf(LL_ERROR, "Protocol not HTTP or malformed HTTP request line in header: '%s'.", 
					buffer);
				ok = false;
				*status = t;
				break;
			}
			header->request_line = strdup(buffer);
		}
		//status line
		else if(i == 0 && type == HEADER_TYPE_REPLY) {
			if(!parse_status_line(buffer, header)) {
				log_printf(LL_ERROR, "Protocol not HTTP or malformed HTTP status line in header: '%s'.", buffer);
				ok = false;
				//status not relevant with replys
				break;
			}
			header->status_line = strdup(buffer);
			//log_printf(LL_DEBUG, "status_code='%s'\n", header->status_code);
		}
		else if(length > 0) {
		
			//header line
			if(!parse_header_line(buffer, &name, &value)) {
				log_printf(LL_ERROR, "Malformed header line received: '%s'.", buffer);
				ok = false;
				*status = STATUS_SYNTAX_ERROR_IN_HEADER_LINE;
				break;
			}
			
			//log_printf(LL_DEBUG, "name='" PSTRING_FORMAT "', value='" PSTRING_FORMAT "'",
			//	PSTRING_ARG(name), PSTRING_ARG(value));
				
			//field names are case insensitive
			if(pstring_iequals(&name, "Content-Type")) {
				header->content_type = pstring_to_cstring(&value);
			} else if(pstring_iequals(&name, "Content-Length")) {
				errno = 0;
				header->content_length = strtoull(value.ptr, NULL, 10);
				//log_printf(LL_DEBUG, "content_length=%llu", header->content_length);
				header->content_length_valid = errno == 0;					
			} else if(pstring_iequals(&name, "Iam")) {
				header->iam = pstring_to_cstring(&value);
				log_printf(LL_INFO, "Iam name: '%s'", header->iam);
			} /* else if(pstring_iequals(&name, "Connection") 
				&& pstring_iequals(&value, "close")) {
				header->close_connection = true;
			} */
			//ignore other header lines
		}
		
		i++;
	} while(length > 0);

	//parsing did not succeed, clean up
	if(!ok) {
		free_header(header);
		header = NULL;
	}

	return header;
}

bool free_header(struct HEADER *header)
{

	if(header == NULL)
		return false;
	
	if(header->request_line != NULL)
		safer_free((char *)header->request_line);
	
	if(header->request_path != NULL)
		safer_free((char *)header->request_path);
	
	if(header->status_line != NULL)
		safer_free((char *)header->status_line);
	
	if(header->status_code != NULL)
		safer_free((char *)header->status_code);
	
	if(header->content_type != NULL)
		safer_free((char *)header->content_type);
	
	if(header->iam != NULL)
		safer_free((char *)header->iam);

	safer_free(header);

	return true;
}

/**
 * Convert a hexadecimal digit in ASCII representation to its value. If the
 * character is not a valid hexadecimal digit, -1 is returned.
 *
 * @param c The character to convert.
 * @return Returns the value of the hexadecimal digit on sucess, -1 on failure.
 */
int hexdigit(char c)
{
	int value = -1;
	c = toupper(c);

	if(isdigit(c))
		value = c - '0';
	else if(c >= 'A' && c <= 'F')
		value = c - 'A' + 10;
	
	return value;
}

/**
 * Decode the %XY escape sequences in an absolute path. See the documentation
 * for a state diagram of the decoding FSM.
 *
 * @param request_path The request path to decode. The original string is 
 *   overwritten.
 * @return Returns true on success, false on failure.
 */
bool decode_request_path(char *request_path)
{
	//reading from s, writing to t
	char *s, *t;
	enum {NORMAL, ESCAPE_1, ESCAPE_2} state;
	int c;
	
	if(request_path == NULL)
		return false;
	
	state = NORMAL;
	for(s = t = request_path; *s != '\0'; s++) {
		if(state == NORMAL) {
			if(*s == '%')
				state = ESCAPE_1;
			else
				*t++ = *s;
		} else if(state == ESCAPE_1) {
			if(isxdigit(*s)) {
				c = hexdigit(*s) << 4;
				state = ESCAPE_2;
			}
			else
				break;
		} else if(state == ESCAPE_2) {
			if(isxdigit(*s)) {
				c |= hexdigit(*s);
				*t++ = c;
				state = NORMAL;
			} else
				break;
		} else {
			break;
		}
	}
	
	//set end of string
	*t = '\0';
	
	return state == NORMAL;
}

/**
 * Validate a request path. The request path is server specific and is not
 * specified in the HTTP specs. From the HTTP perspective, the request path is
 * only an opaque identifyer and has no structure. It does not have to 
 * correspond to any file system structure. However, in this implementation,
 * the request path corresponds to the file system structure relative to the 
 * root path (given as a --root argument). The request path begins with one
 * slash (/). The rest of the path is divided by slashes (/) in to segments.
 * Only one consecutive slash is allowed. No path segment can be . or ..
 * for safety reasons, so that a maliciously formed path would access files
 * outside the root directory. Directories have syntactically valid paths.
 *
 * Examples of syntactically valid paths:
 * /, /file, /dir/file, /dir1/dir2/file, /dir, /dir1/dir2 /dir/
 * Examples of syntactically invalid paths:
 * ///, /dir/../file, /../.././file, , /., dir/file, file, /dir//file
 *
 * @param request_path The request path to be validated.
 * @return Returns true if request_path is valid, false if it is invalid or on any
 *   failure.
 */
bool validate_request_path(const char *request_path)
{
	const char *s;
	size_t len;
	bool ok = false;
	struct PSTRING ps;
	
	if(request_path == NULL)
		return false;
	
	for(s = request_path; *s != '\0'; ) {
		ok = false;

		//one consecutive slash allowed
 		len = strspn(s, "/");
 		if(len != 1)
 			break;
 		s += len;
		
		len = strcspn(s, "/");
		pstring_init(&ps, s, len);
		
		//. and .. not allowed as path segments
		if(pstring_equals(&ps, "."))
			break;
		else if(pstring_equals(&ps, ".."))
			break;

		s += len;
		ok = true;
	}
	
	return ok;
}
