/*
//
// S-38.3610 Network Programming (Spring 2013)
// Assignment 2: HTTP Server
// Filip Nybck (79808K)
//
*/

//standard C headers
#include <stdlib.h>

//unix headers
#include <pthread.h>

//local headers
#include "config.h"
#include "threads.h"

//thread table
struct THREAD *thread_table = NULL;
volatile int thread_count = 0;
//lock for the thread table to prevent concurrent access to the thread table
pthread_mutex_t thread_table_mutex;
//condition variable indicating that the thread table is empty
//=all threads have finished
pthread_cond_t thread_table_empty;
//condition variable indicating that the thread table is no longer full
pthread_cond_t thread_table_not_full;

//lock for handling memory (malloc and free)
pthread_mutex_t memory_mutex;

/**
 * Initialize thread handling. Create mutexes (locks) and condition variables 
 * needed for allocation and release of entries in the thread table. Also 
 * create the memory management mutex (lock).
 *
 * @param config Configuration data to use.
 * @return Returns true on success, false on failure.
 */
bool init_threads(struct CONFIG *config)
{
	int i;
	bool ok = false;

	do {

		thread_table = (struct THREAD*)malloc(config->max_threads*sizeof(struct THREAD));
		if(thread_table == NULL)
			break;

		//initialize the thread table
		for(i = 0; i < config->max_threads; i++)
			thread_table[i].reserved = false;

		//create the thread table lock
		pthread_mutex_init(&thread_table_mutex, NULL);

		//init the thread table condition variables
		pthread_cond_init(&thread_table_empty, NULL);
		pthread_cond_init(&thread_table_not_full, NULL);

		//create the memory handling lock
		pthread_mutex_init(&memory_mutex, NULL);
	
		ok = true;
	} while(false);
	
	return ok;
}

/**
 * Shut down thread handling. First wait that all threads have finished. 
 * Release mutexes (locks) and condition variables
 * 
 * @return Returns true on success, false on failure.
 */
void shutdown_threads()
{
	//acquire the thread table lock
	pthread_mutex_lock(&thread_table_mutex);
	
	//wait that all threads have finished
	while(thread_count > 0) {
		pthread_cond_wait(&thread_table_empty, &thread_table_mutex);
	}

	//release the thread table lock
	pthread_mutex_unlock(&thread_table_mutex);

	//destroy the thread table lock
	pthread_mutex_destroy(&thread_table_mutex);

	//destroy the thread table condition variable
	pthread_cond_destroy(&thread_table_empty);
	pthread_cond_destroy(&thread_table_not_full);

	if(thread_table != NULL) {
		free(thread_table);
		thread_table = NULL;
	}

	//destroy the memory handling lock
	pthread_mutex_destroy(&memory_mutex);
}

/**
 * Allocate an entry in the thread table.
 *
 * @param config Configuration data to use.
 * @return Returns a pointer to the allocated entry in the thread table on
 *   success, NULL on failure.
 */
struct THREAD *alloc_thread_struct(struct CONFIG *config)
{
	struct THREAD *t = NULL;
	int i;

	//acquire the thread table lock
	pthread_mutex_lock(&thread_table_mutex);
	
	//wait until some thread has finished
	while(thread_count == config->max_threads) {
		pthread_cond_wait(&thread_table_not_full, &thread_table_mutex);
	}
	
	//search for a free (unreserved) entry in the thread table
	for(i = 0; i < config->max_threads; i++) {
		if(!thread_table[i].reserved) {
			thread_table[i].reserved = true;
			thread_count++;
			t = &thread_table[i];
			break;
		}
	}
	
	//release the thread table lock
	pthread_mutex_unlock(&thread_table_mutex);

	return t;
}

/**
 * Release a previously allocated entry in the thread table.
 * 
 * @param thread The entry in the thread table to release.
 * @return Returns true on success, false on failure.
 */
bool release_thread_struct(struct THREAD *thread)
{
	if(thread == NULL)
		return false;

	//acquire the thread table lock
	pthread_mutex_lock(&thread_table_mutex);

	thread->reserved = false;
	thread_count--;

	//if this was the last thread, the table is empty
	if(thread_count == 0)
		pthread_cond_signal(&thread_table_empty);
	
	//the thread table is at least not full
	pthread_cond_signal(&thread_table_not_full);

	//release the thread table lock
	pthread_mutex_unlock(&thread_table_mutex);
	
	return true;
}

/**
 * A thread safer variant of malloc protected by a mutex (lock) to prevent 
 * concurrent allocation and freeing of memory.
 *
 * @param size The size of the memory block to allocate.
 * @return Returns a pointer to the allocated memory block on success, NULL on
 *   failure.
 */
void *safer_malloc(size_t size)
{
	void *ptr;

	//acquire the memory handling lock
	pthread_mutex_lock(&memory_mutex);

	ptr = malloc(size);

	//release the memory handling lock
	pthread_mutex_unlock(&memory_mutex);
	
	return ptr;
}

/**
 * A thread safer variant of free protected by a mutex (lock) to prevent 
 * concurrent allocation and freeing of memory.
 *
 * @param ptr The pointer to the previously allocated memory block.
 */
void safer_free(void *ptr)
{
	//acquire the memory handling lock
	pthread_mutex_lock(&memory_mutex);

	free(ptr);

	//release the memory handling lock
	pthread_mutex_unlock(&memory_mutex);
}

